from .sys_utils import show_verbose

from urllib.parse import urlparse
from ipaddress import ip_address

from requests_futures.sessions import FuturesSession
import socket
import requests
import time
import logging

from requests.exceptions import RequestException

from typing import Optional


def query_external_ip(cache=[], timeout=5):
    if cache:
        assert len(cache) == 1
        return cache[0]

    # try to use external ip address, if it fails, use the local ip address
    # failures could be due to several reasons, e.g., enterprise gateway
    try:
        ip = requests.get('https://api.ipify.org', timeout=5).text
        # store external ip because the query takes time
        cache.append(ip)
        return ip
    except RequestException:
        return socket.gethostbyname(socket.gethostname())


def query_ip(url: Optional[str] = None):
    if url:
        rst = urlparse(url)
        hostname = rst.netloc if rst.netloc else url
        if len(hostname) == 0:
            raise ValueError(f"invalid url {url}")
        try:
            ip = socket.gethostbyname(hostname)
        except socket.gaierror:
            return "0.0.0.0"
    else:
        ip = query_external_ip()

    try:
        ip_address(ip)
    except ValueError:
        ip = '0.0.0.0'
    return ip


def port_response_time(host, port, timeout=2):
    """
    return the network latency to host:port, if the latency exceeds
    timeout then return timeout.

    If host is '0.0.0.0' then directly return a number larger than timeout.
    """
    if host == '0.0.0.0':
        return 10*timeout

    start = time.time()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    result = sock.connect_ex((host, port))
    roundtrip = time.time() - start
    return min(roundtrip, timeout)


def first_response(url_lists, timeout):
    """
    send HEAD request to each url, return the first url that responses and skip drop all other urls.
    """
    def _query(url):
        if show_verbose():
            print(f"send HEAD request to {url}")
        return session.head(url, timeout=timeout, allow_redirects=True)

    if show_verbose():
        print(f"HEAD request timeout: {timeout}")

    with FuturesSession(max_workers=5) as session:
        futures = [_query(url) for url in url_lists]

        while True:
            time.sleep(0.1)

            status = [x.running() for x in futures]
            for future in futures:
                try:
                    if future.done() and future.result().status_code == 200:
                        return future.result().url
                except RequestException:
                    continue

            if not any(status):
                return None



from .filters import generate_info
from .gpg_utils import verify_gpg
from .interactive_utils import query_yes_no
from .interactive_utils import color
from .mount_utils import TarMounter, DmgMounter
from .sys_utils import current_architecture, current_system, current_libc
from .sys_utils import show_verbose
from .version_utils import Version
from .version_utils import latest_version
from .version_utils import is_version_released
from .version_utils import is_full_version
from .version_utils import read_releases
from .source_utils import SourceRegistry
from .source_utils import show_upstream
from .source_utils import verify_upstream

__all__ = [
    # filters
    "generate_info",

    # gpg_utils
    "verify_gpg",

    # interactive_utils
    "query_yes_no",
    "color",

    # mount_utils
    "TarMounter", "DmgMounter",

    # sys_utils
    "current_architecture",
    "current_system",
    "current_libc",
    "show_verbose",

    # version_utils
    "Version",
    "is_full_version",
    "latest_version",
    "is_version_released",
    "read_releases",

    # source_utils
    "SourceRegistry",
    "show_upstream",
    "verify_upstream"
]



from .defaults import GPG_PUBLIC_KEY_PATH

from gnupg import GPG
from tempfile import TemporaryDirectory


def _verify_gpg(gpg: GPG, datafile, signature_file):
    # this requires gnupg installed in the system
    with open(signature_file, "rb") as fh:
        return gpg.verify_file(fh, data_filename=datafile)


def verify_gpg(datafile, signature_file=None) -> bool:
    """
    verify Julia releases using GPG
    """
    if signature_file is None:
        signature_file = datafile + ".asc"

    with open(GPG_PUBLIC_KEY_PATH) as fh:
        keycontent = fh.read()

    rst = False
    try:
        with TemporaryDirectory() as tmpdir:
            gpg = GPG(gnupghome=tmpdir)
            gpg.import_keys(keycontent)
            rst = bool(_verify_gpg(gpg, datafile, signature_file))
    except FileNotFoundError:
        # issue #45 -- might reaches here during the temp dir cleanup
        # This might only related to VM environment
        # https://bugs.python.org/issue25717
        pass
    finally:
        return rst



from tempfile import mkdtemp
import os
import sys
import subprocess
import shutil
import time


class Mounter:
    def __init__(self, src_path, mount_root="."):
        self.src_path = src_path
        self.mount_root = os.path.abspath(mount_root)
        mount_name = os.path.splitext(os.path.split(self.src_path)[1])[0]
        self.mount_point = os.path.join(self.mount_root, mount_name)


class TarMounter(Mounter):
    def __enter__(self):
        self.tempdir = mkdtemp()
        # this only supports compressed tarball: *.tar.gz and *.tgz
        args = ["tar", "-zxf", self.src_path, "-C", self.tempdir]
        extra_args = ["--strip-components", "1"]

        args.extend(extra_args)
        is_success = subprocess.run(args).returncode == 0
        if is_success:
            return self.tempdir
        raise IOError(f"could not untar {self.src_path}")

    def __exit__(self, type, value, tb):
        shutil.rmtree(self.tempdir)


class DmgMounter(Mounter):
    def __init__(self, src_path, mount_root=".", verbose=False, max_try=5):
        super(DmgMounter, self).__init__(src_path, mount_root)
        self.extra_args = ["-mount", "required"]
        self.max_try = max_try
        if not verbose:
            self.extra_args.append("-quiet")

    @staticmethod
    def umount(mount_point):
        if os.path.exists(mount_point):
            rst = subprocess.run(["umount", mount_point])
            return not rst.returncode
        return True

    def __enter__(self):
        assert sys.platform == "darwin"
        args = ["hdiutil", "attach", self.src_path,
                "-mountpoint", self.mount_point]
        args.extend(self.extra_args)
        DmgMounter.umount(self.mount_point)

        # the mount might fail for unknown reason,
        # set a max_try here to work it around
        cur_try = 1
        while cur_try <= self.max_try:
            is_success = subprocess.run(args).returncode == 0
            if is_success:
                return self.mount_point
            time.sleep(0.5)
            cur_try += 1

        raise IOError(f"{self.src_path} is not mounted successfully")

    def __exit__(self, type, value, tb):
        DmgMounter.umount(self.mount_point)



from .sys_utils import current_system
import os
import json

from typing import Dict

# this file isn't really a python script
# just some configuration constants
PKG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                        ".."))


def get_configfiles(filename):
    """generate a list of config files ordered with their priorities"""
    configfile_list = []
    sys = current_system()
    if sys in ["mac", "freebsd", "linux"]:
        configfile_list.append(
            os.path.join(os.path.expanduser("~"),
                         ".config", "jill", filename)
        )
    elif sys == "winnt":
        configfile_list.append(
            os.path.join(os.path.expanduser(r"~\AppData\Local\julias"),
                         filename)
        )

    # fallback config
    configfile_list.append(os.path.join(PKG_ROOT, "config", filename))
    return configfile_list


SOURCE_CONFIGFILE = get_configfiles("sources.json")
GPG_PUBLIC_KEY_PATH = os.path.join(PKG_ROOT, ".gnupg", "juliareleases.asc")
DEFAULT_VERSIONS_URL = "https://julialang-s3.julialang.org/bin/versions.json"
VERSIONS_SCHEMA_URL = "https://julialang-s3.julialang.org/bin/versions-schema.json"

# for mirror usage: where releases are downloaded to
default_path_template = "releases/$vminor_version/$filename"


def load_versions_schema(download=False, cache=dict()) -> Dict:
    """
        Load the schema configs.

        The schema is used to validate if the downloaded `versions.json` are valid.
    """
    configfile = os.path.join(PKG_ROOT, "config", "versions-schema.json")
    # Avoid unnecessary IO reads by caching the content into memeory
    if not cache:
        with open(configfile, "r") as file:
            schema = json.load(file)
        cache.update(schema)
    return cache


def load_alias(cache=dict()) -> Dict:
    """
        Load the alias configs.

        For better user experiences, jill allows alias for some os/arch, e.g., `windows`->`winnt`.
    """
    configfile = os.path.join(PKG_ROOT, "config", "alias.json")
    # Avoid unnecessary IO reads by caching the content into memeory
    if not cache:
        with open(configfile, "r") as file:
            alias = json.load(file)
        cache.update(alias)
    return cache


def load_placeholder(
        cache=dict()) -> Dict:
    """
        Load the placeholder configs.

        Placeholders are used to generate URLs from given template, see also "sources.json"
        for an example.
    """
    # Avoid unnecessary IO reads by caching the content into memeory
    configfile = os.path.join(PKG_ROOT, "config", "placeholders.json")
    if not cache:
        with open(configfile, "r") as file:
            schema = json.load(file)
        cache.update(schema)
    return cache


default_filename_template = "julia-$version-$osarch.$extension"
default_latest_filename_template = "julia-latest-$osbit.$extension"

# ports
default_scheme_ports = {
    "rsync": 873,
    "https": 443,
    "http": 80,
    "ftp": 21,
}



from .download import download_package
from .install import install_julia
from .utils import show_upstream
from .mirror import mirror
import fire
import logging
import os


def main():
    logging.basicConfig(filename=os.devnull,
                        level=logging.DEBUG)

    fire.Fire({
        'download': download_package,
        'install': install_julia,
        'upstream': show_upstream,
        'mirror': mirror
    }, name="jill")


if __name__ == "__main__":
    main()



from .filters import f_major_version, f_minor_version, f_patch_version
from .filters import canonicalize_arch, canonicalize_sys
from .defaults import load_versions_schema
from .defaults import DEFAULT_VERSIONS_URL, VERSIONS_SCHEMA_URL
from .net_utils import first_response
from .source_utils import read_registry
from .interactive_utils import color
import semantic_version

import requests
import json
import jsonschema

from jsonschema.exceptions import ValidationError
from typing import Tuple, List


def is_full_version(version: str):
    if version == "latest":
        return True
    try:
        semantic_version.Version(version)
    except ValueError:
        return False
    return True


class Version(semantic_version.Version):
    """
    a thin wrapper on semantic_version.Version that

    * accepts "latest"
    * accepts partial version, e.g., `1`, `1.0`
    """

    def __init__(self, version_string):
        if version_string == "latest":
            version_string = "999.999.999"
            self.major_version = "latest"
            self.minor_version = "latest"
            self.patch_version = "latest"
        else:
            version_string = Version.get_full_version(version_string)
            self.major_version = f_major_version(version_string)
            self.minor_version = f_minor_version(version_string)
            self.patch_version = f_patch_version(version_string)
        super(Version, self).__init__(version_string)

    # TODO: we can actually wrap latest_version here
    @staticmethod
    def get_full_version(version: str):
        """
        add trailing 0s for incomplete version string
        """
        splited = str(version).split(".")
        if len(splited) == 1:
            major, minor, patch = splited[0], "0", "0"
        elif len(splited) == 2:
            major, minor = splited[0:2]
            patch = "0"
        elif len(splited) == 3:
            major, minor, patch = splited[0:3]
        else:
            raise ValueError(f"Unrecognized version string {version}")
        return ".".join([major, minor, patch])


def read_remote_versions(upstream, cache=dict()):
    """
        If not cached, download and read json content from remote URL `url`.
    """
    if not cache:
        registry = read_registry()
        if upstream in registry and registry[upstream].versions_url is not None:
            # For backward-compatibility, `versions` item is optional (issue #64)
            versions_url = registry[upstream].versions_url
        else:
            # If not specified, use the first response as the result
            # TODO: we can share the first_response result here so that later
            #       downloading (e.g., `query_download_url`) does not need to check
            #       it again.
            versions_url_list = [x.versions_url for x in registry.values()]
            versions_url = first_response(versions_url_list, timeout=0.5)
            versions_url = versions_url if versions_url else DEFAULT_VERSIONS_URL
        print(
            f'{color.GREEN}querying release information from {versions_url}{color.END}')
        version_list = json.loads(requests.get(versions_url).text)

        # Validate the downloaded content with `versions_schema.json`.
        # This file is unlikely to be outdated so we keep a copy
        # inside `jill`. If it gets outdated, then print a warning message and
        # download the lastest schema file.
        # When there're new arch/sys that makes this our copy outdated, it's very
        # likely that `jill` doesn't support it.
        schema = load_versions_schema()
        try:
            jsonschema.validate(version_list, schema)
            is_valid = True
        except ValidationError:
            is_valid = False
            print(
                f'{color.YELLOW} failed to validate versions file, retry with latest schema...')
            schema = json.load(requests.get(VERSIONS_SCHEMA_URL).text)
        if not is_valid:
            jsonschema.validate(version_list, schema)

        cache.update(version_list)
    return cache


def read_releases(minimal_version='0.6.0', stable_only=False, upstream=None) -> List[Tuple[str, str, str]]:
    """
    read release info from versions.json.
    The content will be cached so will only download the data once.
    """
    # Here we only cache the raw content so that post processing like `stable_only` filter
    # works as expected.
    v = read_remote_versions(upstream=upstream)
    releases = []
    for item in v.items():
        ver = item[0]
        is_stable = item[1]['stable']
        if not stable_only or is_stable:

            # minimal_version works only when is_stable=True
            try:
                # TODO: a cleaner solution
                if Version(ver) < Version(minimal_version):
                    continue
            except:
                continue

            files = item[1]['files']
            for file in files:
                os = file['os']
                libc = file['triplet'].split('-')[2]
                if libc == "musl":
                    # currently Julia tags musl as a system, e.g.,
                    # https://julialang-s3.julialang.org/bin/musl/x64/1.5/julia-1.5.1-musl-x86_64.tar.gz
                    os = "musl"
                releases.append((ver, os, file['arch']))
    return releases


def is_version_released(version, system, arch, **kwargs):
    """
        Checks if the given version number is released for the given system and architecture.
        Note: returns True for version="latest" if system and architecture are valid.
    """
    # special case version="latest"
    if version == "latest":
        # All the new stuff in nightly builds might break the schema check, thus we directly
        # return True here without any extra checks.
        return True

    version_list = read_releases(**kwargs)
    item = str(version), canonicalize_sys(system), canonicalize_arch(arch)
    return (item in version_list)


def latest_version(version: str, system, arch, **kwargs) -> str:
    """
    Autocompletes a partial semantic version string to the latest compatible full version string.
    Directly returns `version` if it's already a complete version string (without checking that it's
    a valid one).
    """
    system, arch = canonicalize_sys(system), canonicalize_arch(arch)
    # supporting legacy versions is really of low priority
    if version and Version(version) < Version("0.6.0"):
        raise(ValueError('Julia < v"0.6.0" is not supported.'))

    # Download whatever the user requests; ignores `stable_only`.
    if is_full_version(version):
        return version

    # query system/architecture-compatible releases
    releases = read_releases(**kwargs)
    compat_releases = [x for x in releases if x[1] == system and x[2] == arch]
    if len(compat_releases) == 0:
        raise(ValueError(
            f'Julia release for system {system} and architecture {arch} is not available.'))

    if len(version.strip()) == 0:
        # version is an empty string => try to find latest compatible release
        return max(compat_releases, key=lambda x: Version(x[0]))[0]
    else:
        filtered_compat = [
            x for x in compat_releases if x[0].startswith(version)]
        if len(filtered_compat) == 0:
            latest_ver = max(compat_releases, key=lambda x: Version(x[0]))[0]
            print(f'{color.RED}failed to find latest Julia version for "{version}", "{system}" and "{arch}". Trying latest compatible version "{latest_ver}" instead.{color.END}')
            print(f'{color.RED}you can add `--unstable` flag to also include unstable releases.{color.END}')
            # This is equivalent to:
            # latest_version('', system, arch, stable_only=stable_only)
            return latest_ver
        else:
            latest_compat = max(filtered_compat, key=lambda x: Version(x[0]))
            return latest_compat[0]



"""
tools to detect current system and architecture so that things work with tools
in the `filters` module
"""
import subprocess
import os
import platform
import warnings


def current_system():
    rst = platform.system()
    if rst.lower() == "linux":
        return "linux"
    elif rst.lower() == "darwin":
        return "mac"
    elif rst.lower() == "freebsd":
        return "freebsd"
    elif rst.lower() == "windows":
        return "winnt"
    else:
        raise ValueError(f"Unsupported system {rst}")


def current_architecture():
    arch = platform.machine()
    if arch.lower() == "aarch64":
        return "aarch64"
    elif arch.lower() == "armv7l":
        return "armv7l"
    elif arch.lower() == "i386":
        return "i686"
    elif arch.lower() == "amd64":
        return "x86_64"
    else:
        return arch


def current_libc():
    """
        current_libc()

    Detect which libc is used and return either "glibc" or "musl"
    """
    sys = current_system()
    if sys in ["linux", "freebsd"]:
        libc, _ = platform.libc_ver()
        if not libc:
            # in case glibc is not detected, we manually check the result of `ldd --version`
            # Reference:
            # https://unix.stackexchange.com/questions/120380/what-c-library-version-does-my-system-use
            ldd_call = subprocess.run(
                ["ldd", "--version"], capture_output=True)
            # musl `ldd --version` exits with 1
            version_string = ldd_call.stdout if ldd_call.returncode == 0 else ldd_call.stderr
            version_string = version_string.decode(
                "utf-8").splitlines()[0].lower()
            if "musl" in version_string:
                libc = "musl"
            elif "glibc" in version_string:
                libc = "glibc"
            else:
                warnings.warn(
                    f"failed to read libc version from {version_string}, use glibc as fallback")
                libc = "glibc"
        return libc
    else:
        raise ValueError(f"Unsupported system {sys}")


def show_verbose():
    if "GITHUB_ACTIONS" in os.environ:
        return True

    return os.environ.get("DEBUG", False)



import sys


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


class color:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'
